%
% Trust region minimization with DogLeg trusted step.
%
%  USAGE:
%
%  x = DogLeg( fun, Grad, Hess, x0, epsi )
%
%  or
%
% [x,g,radius] = DogLeg( fun, Grad, Hess, x0, epsi )
%
function varargout = DogLeg( fun, Grad, Hess, x0, epsi )
  beta0  = 1E-4 ;
  beta1  = 0.25  ;
  beta2  = 0.99 ;

  x      = x0 ;
  f      = feval( fun,  x ) ;
  g      = feval( Grad, x ) ;
  H      = feval( Hess, x ) ;
  radius = norm(H\g)/2 ;

  if nargout == 3
    varargout{1} = x ;
    varargout{2} = norm(g,inf) ;
    varargout{3} = radius ;
  end ;

  for k=1:1000

    disp(sprintf('iter=%d radius=%f fun=%f',k,radius,f));

    % compute model
    g = feval( Grad, x ) ;
    H = feval( Hess, x ) ;
    
    done = false ;
    
    while ~done
      % compute the step
      s    = DogLegStep( radius, g, H ) ;
      xnew = x + s ;
      fnew = feval( fun, xnew ) ;

      if norm(s,inf) < epsi & norm(g,inf) < epsi
        disp('solution found!');
        return ;
      end;
    
      % check the reduction
      ared  = f - fnew ; 
      pred  = - dot( g, s )  - 0.5 * dot(s,H*s) ;
      rho   = ared / pred ;
      
      if rho >= beta2 % very successfull
        done   = true ;
        x      = xnew ;
        f      = fnew ;
        radius = max(radius, 2*norm(s)) ;
      elseif rho >= beta1 % successful 
        done   = true ;
        x      = xnew ;
        f      = fnew ;
      else
        disp('reject') ;
        radius = radius / 2 ;
      end
    end

    if nargout == 3
      varargout{1} = [ varargout{1}, x ] ;
      varargout{2} = [ varargout{2}, norm(g,inf) ] ;
      varargout{3} = [ varargout{3}, radius ] ;
    end

  end ;

  if nargout == 1
    varargout{1} = x ;
  end ;

end

function s = DogLegStep( radius, g, H )
  mu = 0 ;
  Sg = -g*(dot(g,g)/dot(g,H*g)) ;
  Sn = -H\g ;
  a  = dot(Sg,Sg) ;
  b  = dot(Sn,Sn) ;
  
  if a >= radius^2 
    s = (radius/sqrt(a))*Sg ;
  elseif b > radius^2
    c     = dot( Sg - Sn, Sg - Sn );
    d     = (a+b-c) / 2 ;
    alpha = (b-radius^2)/(b-d+sqrt(d^2-a*b+c*radius^2)) ;
    s     = alpha * Sg + (1-alpha) * Sn ;
  else
    s = Sn ;    
  end
end
