%
% Newton method for non linear system with damping
%
function varargout = NewtonDamped( F, JF, x0, epsi, varargin )
  global NewtonDumped_global ;
  NewtonDumped_global . F  = F ;
  NewtonDumped_global . JF = JF ;

  linesearch = 'wolfe' ;
  if nargin >= 5 ; linesearch = varargin{1} ; end ;

  % if required save the computed iteration
  if nargout == 2
    varargout{1} = [] ;
    varargout{2} = [] ;
  end ;

  x = x0 ;

  for k=1:1000

    % if required save the computed iteration
    if nargout == 2
      varargout{1} = [varargout{1}, norm(feval( F, x ),inf)]  ;
      varargout{2} = [varargout{2}, x ] ;
    end ;

    % compute search direction
    Fx  = feval( F,  x ) ;
    JFx = feval( JF, x ) ;
    
    % check for bad termination
    rc = rcond(JFx) ;
    if ~ isfinite(rc) | rc < 1E-10
      disp('NewtonDamped: bad termination');
      return ;
    end ;

    % compute search direction
    d = -JFx \ Fx ;

    % Setup initial point and direction for linesearch
    NewtonDumped_global . x0 = x ;
    NewtonDumped_global . d  = d ;
    NewtonDumped_global . J  = JFx ;
    
    % perform selected linesearch
    switch lower(linesearch)
      case 'wolfe'  ; alpha = WolfeLineSearch( @Fdir, @DFdir ) ;
      case 'armijo' ; alpha = ArmijoLineSearch( @Fdir, @DFdir) ;
      case 'simple' ; alpha = ArmijoSimpleSearch( @Fdir, @DFdir ) ;
      otherwise ; error( ['NonlinearCG bad linesearch: ' linesearch] ) ;
    end

    % update x
    x = x + alpha * d ;
    disp([k alpha])

    % check for termination
    if norm( d, inf ) < epsi
      break ;
    end ;

  end ;

  if nargout == 1
    varargout{1} = x ;
  end ;

end

function res = Fdir( t )
  global NewtonDumped_global ;
  F   = NewtonDumped_global . F ;
  J   = NewtonDumped_global . J ;
  x0  = NewtonDumped_global . x0 ;
  d   = NewtonDumped_global . d ;
  FF  = J\feval(F, x0 + t * d) ;
  res = min( norm(FF, inf),norm(FF, 1))^2 ;
end

function res = DFdir( t )
  h   = 1E-4 ;
  res = (Fdir(t+h)-Fdir(t-h))/(2*h) ;
end
