%
% Fletcher and Reeves and variants non linear minimization
%
% If there is only one output argument x it will contains the last iterate.
% For example:
% x = NonlinearCG( MODEL, x0, epsi )
%
% If there are two output argument x they will contains the history of the
% iterate, both x and f(x).
% For example:
% [fs, xs] = NonlinearCG( MODEL, x0, epsi )
%
function varargout = NonlinearCG( fun, Grad, x0, epsi, varargin )

  global NonlinearCG_global ;
  NonlinearCG_global . fun  = fun ;
  NonlinearCG_global . Grad = Grad ;
  
  x = x0 ;
  g = feval( Grad, x ) ;
  p = -g ;

  if nargout == 2
    varargout{1} = feval( fun, x )  ;
    varargout{2} = x ;
  else
    varargout{1} = x0 ;
  end ;

  cg         = 'FR' ;
  linesearch = 'wolfe' ;

  if nargin >= 5 ; cg         = varargin{1} ; end ;
  if nargin >= 6 ; linesearch = varargin{2} ; end ;

  gg = dot(g,g);

  for k=1:1000
      
    disp(k)

    % if required save the computed iteration
    if nargout == 2
      varargout{1} = [ varargout{1}, feval( fun, x ) ] ;
      varargout{2} = [ varargout{2}, x ] ;
    end ;

    % check for termination
    if norm( g, inf ) < epsi
      break ;
    end ;

    % Setup initial point and direction for linesearch
    NonlinearCG_global . x0 = x ;
    NonlinearCG_global . d  = p ;
    
    % perform selected linesearch
    switch lower(linesearch)
      case 'wolfe'  ; alpha = WolfeLineSearch( @Fdir, @DFdir ) ;
      case 'armijo' ; alpha = ArmijoLineSearch( @Fdir, @DFdir ) ;
      case 'simple' ; alpha = ArmijoSimpleSearch( @Fdir, @DFdir ) ;
      otherwise ; error( ['NonlinearCG_global bad linesearch: ' linesearch] ) ;
    end
    
    s  = alpha * p ;
    x  = x + s ;
    gn = feval( Grad, x ) ;
    y  = gn - g ;

    % Residual orthogonalization
    switch cg
      case 'HS'  ; % Hestenes and Stiefel
        beta = dot(y,gn)/dot(y,p) ;
      case 'PRP' ;  % Polak Ribiere Polyak
        beta = dot(y,gn)/dot(g,g) ;
      case 'LS'  ; % Liu and Storey
        beta = -dot(y,gn)/dot(g,p) ;
      case 'FR'  ; % Fletcher and Reeves
        beta = dot(gn,gn) / dot(g,g) ;
      case 'CD'  ; % Conjugate Descent - Fletcher
        beta = -dot(gn,gn) / dot(g,p) ;
      case 'DY'  ; % Dai and Yuan
        beta = dot(gn,gn) / dot(y,p) ;
      case 'HuS' ; % Hu and Storey
        beta = max(0,min(dot(y,gn),dot(gn,gn))) / dot(g,g);
      case 'TaS' ; % Touati-Ahmed and Storey
        prp = dot(y,gn)  / dot(g,g) ;
        fr  = dot(gn,gn) / dot(g,g) ;
        if 0 <= prp & prp <= fr
          beta = prp ;
        else
          beta = fr ;
        end ;
      case 'LS-CD' ; % Hybrid LS-CD
        beta = max(0,-max(dot(y,gn)/dot(g,p),dot(gn,gn)/dot(g,p))) ;
      case 'gradient' ; % gradient method
        beta = 0 ;
      otherwise ; error( ['NonlinearCG_global bad method: ' cg] ) ;
    end ;

    % check for restart
    if abs( dot(g,gn) ) > 0.2*dot(gn,gn)
      beta = 0 ;
    end ;
    
    % update search direction
    p  = beta * p - gn ;

    % save for next iteration
    g  = gn ;
    
    % check for bad termination
    if ~ all( isfinite(p) )
      disp('NonlinearCG: bad termination');
      break ;
    end ;

  end ;

  if nargout == 1
    varargout{1} = x ;
  end ;

end

function res = Fdir( t )
  global NonlinearCG_global ;
  fun = NonlinearCG_global . fun ;
  x0  = NonlinearCG_global . x0 ;
  d   = NonlinearCG_global . d ;
  res = feval(fun, x0 + t * d ) ;
end

function res = DFdir( t )
  global NonlinearCG_global ;
  Grad = NonlinearCG_global . Grad ;
  x0   = NonlinearCG_global . x0 ;
  d    = NonlinearCG_global . d ;
  res  = dot( feval(Grad, x0 + t * d ), d) ;
end


