%
%
function varargout = QuasiNewton( fun, Grad, x0, epsi, varargin )

  global QuasiNewton_global ;
  QuasiNewton_global . fun  = fun ;
  QuasiNewton_global . Grad = Grad ;

  x = x0 ;
  f = feval( fun,  x ) ;
  g = feval( Grad, x ) ;
  B = eye(length(x0));
  H = eye(length(x0));

  % if required save the computed iteration
  if nargout == 2
    varargout{1} = [] ;
    varargout{2} = [] ;
  end ;

  cg         = 'BFGS' ;
  linesearch = 'wolfe' ;
  if nargin >= 5 ; qn         = varargin{1} ; end ;
  if nargin >= 6 ; linesearch = varargin{2} ; end ;

  for k=1:1000

    % compute search direction
    switch qn
      case {'SR1','DFP','BFGS'} ; d = -H*g ;
      case 'PSB' ;
        % check for bad termination
        if ~ isfinite(rcond(B))
          disp('QuasiNewton: bad termination');
          return ;
        end ;
        d = -B\g ;
      otherwise ; error( ['QuasiNewton bad method: ' qn] ) ;
    end ;

    if nargout == 2
      varargout{1} = [ varargout{1}, feval( fun, x ) ] ;
      varargout{2} = [ varargout{2}, x ] ;
    end ;

    % check for termination
    if norm( g, inf ) < epsi ; break ; end ;    

    % Setup initial point and direction for linesearch
    QuasiNewton_global . x0 = x ;
    QuasiNewton_global . d  = d ;

    % perform selected linesearch
    switch lower(linesearch)
      case 'wolfe'  ; alpha = WolfeLineSearch( @Fdir, @DFdir ) ;
      case 'armijo' ; alpha = ArmijoLineSearch( @Fdir, @DFdir ) ;
      case 'simple' ; alpha = ArmijoSimpleSearch( @Fdir, @DFdir ) ;
      otherwise ; error( ['QuasiNewton_global bad linesearch: ' linesearch] ) ;
    end
    
    disp([k alpha])

    % perform step
    s = alpha * d ;
    x = x + s ;
    
    % update gradient
    ng = feval( Grad, x ) ;
    y  = ng - g ;

    % update H
    % Residual orthogonalization
    switch qn
      case 'SR1'  ; H = SR1  ( H, s, y ) ;
      case 'DFP'  ; H = DFP  ( H, s, y ) ;
      case 'BFGS' ; H = BFGS ( H, s, y ) ;
      case 'PSB'  ; B = PSB  ( B, s, y ) ;
      otherwise ; error( ['QuasiNewton bad method: ' qn] ) ;
    end ;

    g = ng ;

  end;
end

function res = Fdir( t )
  global QuasiNewton_global ;
  fun = QuasiNewton_global . fun ;
  x0  = QuasiNewton_global . x0 ;
  d   = QuasiNewton_global . d ;
  res = feval(fun, x0 + t * d ) ;
end

function res = DFdir( t )
  global QuasiNewton_global ;
  Grad = QuasiNewton_global . Grad ;
  x0   = QuasiNewton_global . x0 ;
  d    = QuasiNewton_global . d ;
  res  = dot(feval(Grad, x0 + t * d ), d) ;
end


function HP = SR1( H, s, y )
  epsi = 0.1 ;
  w    = s - H*y ;
  if dot(w,y) > epsi * norm(w)*norm(y)
    HP = H + (w*w')./dot(w,y);
  else
    HP = H ;
  end
end

function BP = PSB( B, s, y )
  epsi = 0.1 ;
  w    = y - B*s ;
  BP   = B + (w*s'+s*w')./dot(s,s) - (dot(w,s)/dot(s,s)^2) * (s*s') ;
end

function HP = DFP( H, s, y )
  z    = H*y ;
  HP   = H + (s*s')./dot(s,y) - (z*z')./dot(z,y) ;
end

function HP = BFGS( H, s, y )
  beta  = 1/dot(s,y) ;
  z     = beta*(H*y) ;
  alpha = beta+dot(z,y)*beta ;
  HP    = H - (z*s'+s*z') + ((s*alpha)*s') ;
end
